import numpy as np
from astropy import units as u
from astropy import constants as ct
from scipy.interpolate import interp1d, RectBivariateSpline

from .grid import GridTools
from .tools.utils import hypot_func

au_to_m = u.au.to('m')

#*******************
#VELOCITY FUNCTIONS
#*******************
def keplerian(coord, Mstar=1.0, vel_sign=1, vsys=0):
    Mstar *= u.M_sun.to('kg')
    if 'R' not in coord.keys(): R = hypot_func(coord['x'], coord['y'])
    else: R = coord['R'] 
    return vel_sign*np.sqrt(ct.G.value*Mstar/R) * 1e-3

def keplerian_vertical(coord, Mstar=1.0, vel_sign=1, vsys=0):
    Mstar *= u.M_sun.to('kg')
    if 'R' not in coord.keys():
        R = hypot_func(coord['x'], coord['y'])
    else:
        R = coord['R'] 
    if 'r' not in coord.keys():
        r = hypot_func(R, coord['z'])
    else:
        r = coord['r']
    return vel_sign*np.sqrt(ct.G.value*Mstar/r**3)*R * 1e-3 

def velocity_hydro2d(coord, func_interp_R=None, func_interp_phi=None, Mstar=1.0, vel_sign=1, vsys=0, phip=0.0, **dummies):
    Mstar *= u.M_sun.to('kg')
    y = coord['x'] #90 deg shift so that phip=0 aligns with disc major axis
    x = coord['y']
    R = coord['R']

    if phip != 0.0:
        phip = np.radians(phip)
        x, y = GridTools._rotate_sky_plane(x, y, phip)
        
    if 'r' not in coord.keys():
        r = hypot_func(R, coord['z'])
        
    else:
        r = coord['r']

    if func_interp_R is None and func_interp_phi is not None:
        vphi = func_interp_phi(x,y, grid=False)                    
        vR = np.zeros_like(vphi)

    elif func_interp_R is not None and func_interp_phi is None:
        vR = func_interp_R(x,y, grid=False)        
        vphi = np.zeros_like(vR)

    elif func_interp_R is None and func_interp_phi is None:
        return [0, 0, 0]    

    else:
        vR = func_interp_R(x,y, grid=False)
        vphi = func_interp_phi(x,y, grid=False)
        
    #vkep = np.sqrt(ct.G.value*Mstar/r**3)*R * 1e-3
    vz = np.zeros_like(vR)
    
    return [vel_sign*(vphi), vR, 0]

#************
#ORIENTATION
#************
def orientation_constant(coords, incl=np.pi/4, PA=0.0, xc=0.0, yc=0.0):
    xc = xc*au_to_m
    yc = yc*au_to_m
    return incl, PA, xc, yc

#******************
#EMISSION SURFACES
#******************
def z_upper_exp_tapered(coord, z0, p, Rb, q, R0=100):
    R = coord['R']/au_to_m
    return au_to_m*(z0*(R/R0)**p*np.exp(-(R/Rb)**q))

def z_lower_exp_tapered(coord, z0, p, Rb, q, R0=100):
    R = coord['R']/au_to_m
    return -au_to_m*(z0*(R/R0)**p*np.exp(-(R/Rb)**q))

def z_upper_powerlaw(coord, z0, p, Rb, q, R0=100):
    R = coord['R']/au_to_m
    return au_to_m*(z0*(R/R0)**p - Rb*(R/R0)**q)

def z_lower_powerlaw(coord, z0, p, Rb, q, R0=100):
    R = coord['R']/au_to_m
    return -au_to_m*(z0*(R/R0)**p - Rb*(R/R0)**q)

def z_upper_irregular(coord, z0='0.txt', kwargs_interp1d={}, **dummies):
    R = coord['R']/au_to_m
    Rmax = np.nanmax(R)
    Rf, zf = np.loadtxt(z0)
    Rf = np.append(0.0, Rf)
    zf = np.append(0.0, zf)
    if np.max(Rf) < Rmax:
        Rf = np.append(Rf, Rmax)
        zf = np.append(zf, 0.0)
    z_interp = interp1d(Rf, zf, **kwargs_interp1d)
    return au_to_m*z_interp(R)

def z_lower_irregular(coord, z0='0.txt', kwargs_interp1d={}, **dummies):
    R = coord['R']/au_to_m
    Rmax = np.nanmax(R)
    Rf, zf = np.loadtxt(z0)
    Rf = np.append(0.0, Rf)
    zf = np.append(0.0, zf)
    if np.max(Rf) < Rmax:
        Rf = np.append(Rf, Rmax)
        zf = np.append(zf, 0.0)
    z_interp = interp1d(Rf, zf, **kwargs_interp1d)
    return -au_to_m*z_interp(R)

#***************
#PEAK INTENSITY
#***************
def intensity_powerlaw_rout(coord, I0=30.0, R0=100, p=-0.4, z0=100, q=0.3, Rout=500):
    if 'R' not in coord.keys(): R = hypot_func(coord['x'], coord['y'])
    else: R = coord['R']
    z = coord['z']
    R0*=au_to_m
    z0*=au_to_m
    Rout*=au_to_m
    A = I0*R0**-p*z0**-q
    Ieff = np.where(R<=Rout, A*R**p*np.abs(z)**q, 0.0)
    return Ieff

def intensity_powerlaw_rbreak(coord, I0=30.0, p0=-0.4, p1=-0.4, z0=100, q=0.3, Rbreak=20, Rout=500, p=0):
    #p is a dummy variable here
    if 'R' not in coord.keys(): R = hypot_func(coord['x'], coord['y'])
    else: R = coord['R']
    z = coord['z']
    z0*=au_to_m
    Rout*=au_to_m
    Rbreak*=au_to_m
    A = I0*Rbreak**-p0*z0**-q
    B = I0*Rbreak**-p1*z0**-q
    Ieff = np.where(R<=Rbreak, A*R**p0*np.abs(z)**q, B*R**p1*np.abs(z)**q)
    ind = R>Rout
    Ieff[ind] = 0.0
    return Ieff

def intensity_powerlaw_rbreak_nosurf(coord, I0=1.0, p0=-2.5, p1=-1.5,
                                     Rbreak=100, Rout=300, p=0, q=0):
    #p and q are dummy variables here
    if 'R' not in coord.keys(): R = hypot_func(coord['x'], coord['y'])
    else: R = coord['R']
    Rnorm=Rbreak
    Rnorm*=au_to_m
    Rbreak*=au_to_m
    Rout*=au_to_m
    pwl0 =  I0*(R/Rnorm)**p0
    pwl1 =  I0*(R/Rnorm)**p1
    Ieff = np.where(R<=Rbreak, pwl0, pwl1)
    Ieff_rout = np.where(R<=Rout, Ieff, 0.0)
    return Ieff_rout

def intensity_powerlaw_rout_hydro(coord, I0=30.0, R0=100, p=-0.4, z0=100, q=0.3, Rout=500, func_interp_sigma=None, phip=0.0, weight=1.0):
    
    y = coord['x'] 
    x = coord['y']

    if phip != 0.0:
        phip = np.radians(phip)
        x, y = GridTools._rotate_sky_plane(x, y, phip)

    if 'R' not in coord.keys():
        R = hypot_func(coord['x'], coord['y'])
    else:
        R = coord['R']

    if func_interp_sigma is None:
        sigma = np.zeros_like(R)
    else:
        sigma = func_interp_sigma(x,y, grid=False)
        
    z = coord['z']
    R0*=au_to_m
    z0*=au_to_m
    Rout*=au_to_m
    A = I0*R0**-p*z0**-q
    Ieff = np.where(R<=Rout, A*R**p*np.abs(z)**q, 0.0) * sigma**weight

    return Ieff

def intensity_powerlaw_rbreak_hydro(coord, I0=30.0, p0=-0.4, p1=-0.4, z0=100, q=0.3, Rbreak=20, Rout=500, p=0, func_interp_sigma=None, phip=0.0, weight=1.0):
    
    y = coord['x'] 
    x = coord['y']

    if phip != 0.0:
        phip = np.radians(phip)
        x, y = GridTools._rotate_sky_plane(x, y, phip)

    if 'R' not in coord.keys():
        R = hypot_func(coord['x'], coord['y'])
    else:
        R = coord['R']

    if func_interp_sigma is None:
        sigma = np.zeros_like(R)
    else:
        sigma = func_interp_sigma(x,y, grid=False)

    z = coord['z']
    z0*=au_to_m
    Rout*=au_to_m
    Rbreak*=au_to_m
    A = I0*Rbreak**-p0*z0**-q
    B = I0*Rbreak**-p1*z0**-q
    Ieff = np.where(R<=Rbreak, A*R**p0*np.abs(z)**q, B*R**p1*np.abs(z)**q) * (sigma/np.max(sigma))**weight
    #print (np.nanmax(sigma), np.nanmin(sigma))
    ind = R>Rout
    Ieff[ind] = 0.0
    return Ieff

#***********
#LINE WIDTH 
#***********
def linewidth_powerlaw(coord, L0=0.2, p=-0.4, q=0.3, R0=100, z0=100):
    R0*=au_to_m
    z0*=au_to_m
    if 'R' not in coord.keys():
        R = hypot_func(coord['x'], coord['y'])
    else:
        R = coord['R'] 
    z = coord['z']        
    A = L0*R0**-p*z0**-q
    return A*R**p*np.abs(z)**q

#***********
#LINE SLOPE 
#***********
def lineslope_powerlaw(coord, Ls=5.0, p=0.0, q=0.0, R0=100, z0=100):
    R0*=au_to_m
    z0*=au_to_m
    if p==0.0 and q==0.0:
        return Ls
    else:
        if 'R' not in coord.keys():
            R = hypot_func(coord['x'], coord['y'])
        else:
            R = coord['R'] 
        z = coord['z']        
        A = Ls*R0**-p*z0**-q
        return A*R**p*np.abs(z)**q

#**************
#LINE PROFILES
#**************
def line_profile_bell(v_chan, v, v_sigma, b_slope):
    return 1/(1+np.abs((v-v_chan)/v_sigma)**(2*b_slope))        
    
def line_profile_gaussian(v_chan, v, v_sigma, *dummies):
    return np.exp(-0.5*((v-v_chan)/v_sigma)**2)

#***********************
#UPPER + LOWER PROFILES
#***********************
def line_uplow_mask(Iup, Ilow):
    #velocity nans might differ from intensity nans when a z=0 and SG is active, nanmax must be used
    return np.nanmax([Iup, Ilow], axis=0)

def line_uplow_sum(Iup, Ilow):
    return Iup + Ilow    



from discminer.rail import Rail, Contours
from discminer.plottools import (get_discminer_cmap,
                                 make_up_ax, mod_major_ticks,
                                 use_discminer_style, mod_nticks_cbars)

import numpy as np
import matplotlib.pyplot as plt
from astropy import units as u
from astropy.io import fits

import json
import copy
import sys

from argparse import ArgumentParser
from utils import (load_disc_grid,
                   add_parser_args,
                   load_moments,
                   init_data_and_model,
                   get_noise_mask,
                   get_1d_plot_decorators)

use_discminer_style()

parser = ArgumentParser(prog='Plot azimuthal profiles', description='Compute azimuthal profiles from a given moment map [velocity, linewidth, peakintensity]')
parser.add_argument('-t', '--type', default='residuals', type=str,
                    choices=['data', 'model', 'residuals'],
                    help="Compute profiles on data, model or residual moment map. DEFAULTS to 'residuals'")
args = add_parser_args(parser, moment=True, kernel=True, kind=True, surface=True, Rinner=True, Router=True, smooth=True)

#**********************
#JSON AND PARSER STUFF
#**********************
with open('parfile.json') as json_file:
    pars = json.load(json_file)

meta = pars['metadata']
best = pars['best_fit']
custom = pars['custom']

vsys = best['velocity']['vsys']
incl = best['orientation']['incl']
Rout = best['intensity']['Rout']
au_to_m = u.au.to('m')

clabel, clabel_res, clim0, clim0_res, clim1, clim1_res, unit = get_1d_plot_decorators(args.moment, tag=args.type)

if args.type=='residuals':
    clim0 = clim0_res
    clim1 = clim1_res
    clabel = clabel_res

if args.moment=='velocity':
    if args.type=='residuals':
        clabel = r'Velocity %s %s'%(args.type, unit)
    else:
        clabel = r'Deprojected Velocity %s %s'%(args.type, unit)        

#*******************
#LOAD DATA AND MODEL
#*******************
datacube, model = init_data_and_model()

noise_mean, mask = get_noise_mask(datacube, thres=2)
vchannels = datacube.vchannels
model.make_model()

#*************************
#LOAD MOMENT MAPS
moment_data, moment_model, residuals, mtags = load_moments(args, mask=mask)

if args.type=='residuals': map2d = residuals
elif args.type=='data': map2d = moment_data
elif args.type=='model': map2d = moment_model

if args.type!='residuals' and args.moment=='velocity': #deproject velocity field
    map2d = np.abs((map2d-vsys)/(np.cos(model.projected_coords['phi'][args.surface])*np.sin(incl)))

#*************************
r"""
#LOAD DISC GEOMETRY
R, phi, z = load_disc_grid()
RR = R[args.surface]/au_to_m
PP = phi[args.surface]

Xproj = RR*np.cos(phi[args.surface])
Yproj = RR*np.sin(phi[args.surface])

#Read filaments
filaments = fits.getdata('filaments_pos_velocity_gaussian_cartesian.fits')
fp = filaments[2].astype(bool)
pp = np.degrees(PP[fp])
vp = residuals[fp]
xp = Xproj[fp]
yp = Yproj[fp]
#"""

#**************************
#MAKE PLOT

beam_au = datacube.beam_size.to('au').value
R_prof = np.arange(args.Rinner*beam_au, args.Router*Rout, beam_au/5)

color_bounds = np.array([0.33, 0.66, 1.0])*Rout
lws = [0.4, 0.6, 0.4]

rail = Rail(model, map2d, R_prof)

fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(14,6))
ax2 = fig.add_axes([0.85,0.6,0.3*6/14,0.3])

R_ref = 100
R_list, phi_list, resid_list, color_list = rail.prop_along_coords(coord_ref=R_ref, surface=args.surface,
                                                                  color_bounds=color_bounds, lws=lws,
                                                                  ax=ax, ax2=ax2)

tick_angles = np.arange(-150, 180, 30)
ax.set_xticks(tick_angles)
ax.set_xlabel(r'Azimuth [deg]')
ax.set_ylabel(clabel)
ax.set_xlim(-180,180)
ax.set_ylim(clim0, clim1)
ax.grid()

model.make_emission_surface(ax2, which=mtags['surf'])
model.make_disc_axes(ax2)
make_up_ax(ax, labeltop=False)
make_up_ax(ax2, labelbottom=False, labelleft=False, labeltop=True)
ax.tick_params(labelbottom=True, top=True, right=True, which='both', direction='in')

#ax.scatter(pp, vp, edgecolor='k', facecolor='lime', s=40)
#ax2.scatter(xp, yp, edgecolor='k', facecolor='lime', s=10, zorder=20)

plt.savefig('azimuthal_%s_%s.png'%(args.moment, args.type), bbox_inches='tight', dpi=200)
plt.show()
plt.close()






from discminer.mining_control import _mining_moments1d
from discminer.core import Data

from astropy import units as u

import os
import json

if __name__ == '__main__':
    parser = _mining_moments1d(None)
    args = parser.parse_args()

#**************************
#JSON AND SOME DEFINITIONS
#**************************    
with open('parfile.json') as json_file:
    pars = json.load(json_file)

meta = pars['metadata']

dpc = meta['dpc']*u.pc
tag = meta['tag']

#*********
#LOAD DATA
#*********
if args.planck:
    file_data = 'cube_data_%s.fits'%tag
    file_model = 'cube_model_%s.fits'%tag
else:
    file_data = 'cube_data_%s_convtb.fits'%tag
    file_model = 'cube_model_%s_convtb.fits'%tag

if args.fit_data:
    datacube = Data(file_data, dpc) # Read datacube and convert to Cube object
if args.fit_model:
    modelcube = Data(file_model, dpc) # Read modelcube and convert to Cube object
    
#**********************
#MAKE MOMENT MAPS
#**********************
if args.kernel=='quadratic':
    if args.fit_data:
        os.system('bettermoments '+file_data)
        
    if args.fit_model:
        os.system('bettermoments '+file_model)
        
    for old, new in zip(['_v0', '_dv0', '_Fnu', '_dFnu'], ['velocity_', 'delta_velocity_', 'peakintensity_', 'delta_peakintensity_']):
        if args.fit_data:
            os.rename(file_data.split('.fits')[0]+old+'.fits', new+'quadratic_data.fits')
        if args.fit_model:
            os.rename(file_model.split('.fits')[0]+old+'.fits', new+'quadratic_model.fits')
    
else:
    if args.fit_data:
        moments_data = datacube.make_moments(method=args.kernel, tag='data', sigma_thres=args.sigma, peak_kernel=args.peakkernel, fit_continuum=args.fit_continuum)
    if args.fit_model:
        moments_model = modelcube.make_moments(method=args.kernel, tag='model', sigma_thres=0, peak_kernel=args.peakkernel)

#*********************************
#CONVERT TO TB USING PLANCK'S LAW
#*********************************
if args.planck:
    
    filename = 'peakintensity_%s%s.fits'

    if args.fit_data:
        datapeak = Data(filename%(args.kernel, '_data'), dpc, twodim=True)
        datapeak.convert_to_tb(writefits=False, planck=True) #Don't save fits file with custom name
        datapeak.data = datapeak.data.squeeze()
        datapeak.writefits() #Overwrite instead

    if args.fit_model:
        modelpeak = Data(filename%(args.kernel, '_model'), dpc, twodim=True)
        modelpeak.convert_to_tb(writefits=False, planck=True)
        modelpeak.data = modelpeak.data.squeeze()
        modelpeak.writefits()
    



from discminer.mining_control import _mining_pv_diagram
from discminer.mining_utils import (get_2d_plot_decorators,
                                    get_noise_mask,
                                    load_moments,
                                    load_disc_grid,
                                    init_data_and_model,
                                    show_output)

from discminer.rail import Contours
from discminer.grid import GridTools
from discminer.plottools import (make_up_ax,
                                 mod_major_ticks,
                                 mod_nticks_cbars,
                                 get_discminer_cmap,
                                 use_discminer_style)

import json
import copy
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import FormatStrFormatter
from scipy.ndimage import map_coordinates
from astropy import units as u

use_discminer_style()

if __name__ == '__main__':
    parser = _mining_pv_diagram(None)
    args = parser.parse_args()

#**********************
#JSON AND PARSER STUFF
#**********************
with open('parfile.json') as json_file:
    pars = json.load(json_file)

meta = pars['metadata']
best = pars['best_fit']
custom = pars['custom']

Mstar = best['velocity']['Mstar']
vsys = best['velocity']['vsys']
Rout = best['intensity']['Rout']
incl = best['orientation']['incl']
PA = best['orientation']['PA']
xc = best['orientation']['xc']
yc = best['orientation']['yc']
vel_pars = best['velocity']

rings = custom['rings']
gaps = custom['gaps']
kinks = []

ctitle, clabel, clim, cfmt, cmap_mom, cmap_res, levels_im, levels_cc, unit = get_2d_plot_decorators(args.moment)
    
#****************
#SOME DEFINITIONS
#****************
file_data = meta['file_data']
tag = meta['tag']

dpc = meta['dpc']*u.pc
au_to_m = u.au.to('m')

pvphi = np.radians(args.pvphi)
pvcmap = get_discminer_cmap('peakintensity')

#********************
#LOAD DATA AND GRID
#********************
datacube, model = init_data_and_model(Rmin=0, Rmax=1.0)
noise_mean, mask = get_noise_mask(datacube, thres=2)
vchans_shifted = datacube.vchannels - vsys

with open('grid_extent.json') as json_file:
    grid = json.load(json_file)

xmax = grid['xsky'] 
xlim = 1.1*Rout
extent= np.array([-xmax, xmax, -xmax, xmax])

beam_au = datacube.beam_size.to('au').value

if args.surface in ['up', 'upper']:
    z_func = model.z_upper_func
    z_pars = best['height_upper']
    
elif args.surface in ['low', 'lower']:
    z_func = model.z_lower_func
    z_pars = best['height_lower']

#**********************
#DISC REFERENCE SYSTEM
#**********************
R, phi, z = load_disc_grid() #No need to make_model
R_au = R[args.surface]/au_to_m
R_au_flat = R_au.flatten()
phi_rad = phi[args.surface]
phi_rad_flat = phi_rad.flatten()
indices = np.arange(len(R_au_flat))

#*******************************
#MOMENT MAPS AND RADIAL PROFILE
#*******************************
moment_data, moment_model, residuals, mtags = load_moments(args, mask=mask)

R_prof = np.arange(args.Rinner*beam_au, args.Router*Rout, beam_au/4)
R_vkep = np.append(np.linspace(0.1*R_prof[0], R_prof[0], 10), R_prof[1:])

#******
#UTILS
#******
def get_skypixel_ij(Ri, phii, tol=beam_au):

    phi_diff = np.abs(phi_rad - phii).flatten()

    pind = Ri * phi_diff < tol  #pixels where this arc length is smaller than tol    
    rind = np.abs(Ri - R_au_flat[pind]) < tol #same for dR

    R_tol = R_au_flat[pind][rind]
    phi_tol = phi_diff[pind][rind]
    tot_diff = np.hypot(R_tol-Ri, R_tol*phi_tol) #distance between Ri, phii and nearby pixels
    
    indbest = indices[pind][rind][np.argmin(tot_diff)]

    return np.unravel_index(indbest, R_au.shape)

def get_sky_intensity(phii):

    zp = z_func({'R': R_prof*u.au.to('m')}, **z_pars)*u.m.to('au')
    x_samples,y_samples,z_samples = GridTools.get_sky_from_disc_coords(R_prof, phii, zp, incl, PA, xc, yc) 
    
    x_min, x_max, y_min, y_max = extent
    j_samples = (x_samples - x_min) / (x_max - x_min) * (datacube.data.shape[2] - 1)
    i_samples = (y_samples - y_min) / (y_max - y_min) * (datacube.data.shape[1] - 1)

    coords = np.vstack([i_samples, j_samples])
    intensity_along_ray = np.zeros((datacube.nchan, len(R_prof)))
    
    for n in range(datacube.nchan):
        intensity_along_ray[n] = map_coordinates(datacube.data[n], coords, order=1, mode='nearest')
    
    return intensity_along_ray.T    

def get_vkep(Mstar=None):

    R_prof = R_vkep*u.au.to('m')
    z_prof = z_func({'R': R_prof}, **z_pars)

    if Mstar is None:
        v_pars = vel_pars
    else:
        v_pars = copy.copy(vel_pars)
        v_pars['Mstar'] = Mstar
        
    v_m = model.velocity_func({'R': R_prof, 'z': z_prof}, **v_pars)

    return v_m*np.sin(incl)*np.cos(pvphi)

#**********
#MAKE PLOT
#*********
fig, ax = plt.subplots(ncols=2, nrows=1, figsize=(15,8))

ax[0].set_aspect(1)
pos0 = ax[0].get_position()
height1 = 0.5
ax1 = fig.add_axes([pos0.x1+0.05, pos0.y0+0.5*pos0.height-0.5*height1, 0.4, height1])
ax[1].axis('off')

ax_cbar0 = fig.add_axes([pos0.x0, 0.09, pos0.width, 0.04])

intensity_interp_p = get_sky_intensity(pvphi)
intensity_interp_n = get_sky_intensity(pvphi-np.pi)
line_profiles = np.append(intensity_interp_n[::-1], intensity_interp_p, axis=0)

R_axis = np.append(-R_prof[::-1], R_prof)
ax1.contourf(R_axis, vchans_shifted, line_profiles.T, levels=np.linspace(0, np.nanmax(line_profiles), 32), cmap=pvcmap)

ax1.fill_between([-R_prof[0], R_prof[0]], vchans_shifted[0], vchans_shifted[-1], color='k', alpha=0.4)

#***************
#OVERLAY PV CUT
#***************
z_prof = z_func({'R': R_prof*u.au.to('m')}, **z_pars)*u.m.to('au')
isort = np.argsort(R_prof)
 
kwargs_axes = dict(ls=':', lw=4.5, dash_capstyle='round', dashes=(0.5, 1.5), alpha=1)
make_ax = lambda ax, x, y, color: ax.plot(x, y, color=color, **kwargs_axes)        
 
pv_axis0 = np.zeros_like(R_prof) + pvphi
pv_axis1 = np.zeros_like(R_prof) + pvphi - np.pi 
 
x_cont, y_cont, _ = GridTools.get_sky_from_disc_coords(R_prof[isort], pv_axis0, z_prof[isort], incl, PA, xc, yc)    
make_ax(ax[0], x_cont, y_cont, 'tomato')
 
x_cont, y_cont, _ = GridTools.get_sky_from_disc_coords(R_prof[isort], pv_axis1, z_prof[isort], incl, PA, xc, yc)    
make_ax(ax[0], x_cont, y_cont, 'dodgerblue')

#*************
#OVERLAY VKEP
#*************
v_model = get_vkep()
kw_plot = dict(color='k', lw=2, dash_capstyle='round', dashes=(2.0, 2.0), alpha=0.9)
ax1.plot(R_vkep, v_model, -R_vkep, -v_model, **kw_plot)
#v_p = get_vkep(1.1*Mstar)
#v_n = get_vkep(0.9*Mstar)
#ax1.plot(R_vkep, v_p*np.sin(incl), -R_vkep, -v_p*np.sin(incl), **kw_plot)
#ax1.plot(R_vkep, v_n*np.sin(incl), -R_vkep, -v_n*np.sin(incl), **kw_plot)

#***********
#MOMENT MAP
#***********
kwargs_im = dict(cmap=cmap_mom, extent=extent, levels=levels_im)
kwargs_cc = dict(colors='k', linestyles='-', extent=extent, levels=levels_cc, linewidths=0.4)
kwargs_cbar = dict(orientation='horizontal', pad=0.03, shrink=0.95, aspect=15)

im = ax[0].contourf(moment_data, extend='both', **kwargs_im)
if args.surface!='lower':
    cc = ax[0].contour(moment_data, **kwargs_cc)
make_up_ax(ax[0], xlims=(-xlim, xlim), ylims=(-xlim, xlim), labelsize=13, color='k', labelcolor='k')

ax[0].scatter(best['orientation']['xc'], best['orientation']['yc'],
            ec='k', fc='w', marker='X', lw=0.5, s=60, zorder=20)        
datacube.plot_beam(ax[0], fc='0.8')
mod_major_ticks(ax[0], axis='both', nbins=8)

make_up_ax(ax1, xlims=(-R_prof[-1], R_prof[-1]), ylims=(vchans_shifted[0], vchans_shifted[-1]), labelsize=13, color='0.3', labelcolor='0.3', labelbottom=True, labeltop=False, labelright=True)
mod_major_ticks(ax1, axis='both', nbins=8)

for axi in [ax1]:
    axi.tick_params(labelleft=False)
    for side in ['top','bottom','left','right']:
        axi.spines[side].set_linewidth(4.0)
        axi.spines[side].set_color('0.5')
        axi.spines[side].set_linestyle((0, (1,1.5)))
        axi.spines[side].set_capstyle('round')
    axi.grid(color='k', ls='--')

Contours.emission_surface(ax[0], R, phi, extent=extent,
                          R_lev=R_prof*u.au.to('m'),
                          which=mtags['surf'],
                          kwargs_R={'colors': '0.1', 'linewidths': 0.5},
                          kwargs_phi={'colors': '0.1', 'linewidths': 0.4}
)
   


ax[0].set_ylabel('Offset [au]', fontsize=15)
ax[0].set_title(ctitle, pad=40, fontsize=17)

ax1.set_title(r'PV along $\phi=%d^\circ$'%args.pvphi, pad=20, fontsize=17, color='0.3')
ax1.set_xlabel('Offset [au]', color='0.3', fontsize=15)
ax1.set_ylabel(r'$\upsilon_{\rm l.o.s}$ [km/s]', fontsize=16, color='0.3')
ax1.yaxis.set_label_position("right")
ax1.yaxis.set_major_formatter(FormatStrFormatter('%2d'))

cbar0 = plt.colorbar(im, cax=ax_cbar0, format='%.1f', **kwargs_cbar)
cbar0.ax.tick_params(labelsize=12) 
mod_nticks_cbars([cbar0], nbins=6)
cbar0.set_label(clabel, fontsize=15)

plt.savefig('pv_diagram_%s_%ddeg.png'%(mtags['base'], args.pvphi), bbox_inches='tight', dpi=200)
show_output(args)



from discminer.mining_control import _mining_parcube
from discminer.plottools import use_discminer_style
from discminer.tools import fit_kernel
from discminer.core import Data
from discminer.cube import Cube

import numpy as np
import matplotlib.pyplot as plt

from astropy import units as u
from astropy.io import fits

import json

use_discminer_style()

if __name__ == '__main__':
    parser = _mining_parcube(None)
    args = parser.parse_args()

#**************************
#JSON AND SOME DEFINITIONS
#**************************
with open('parfile.json') as json_file:
    pars = json.load(json_file)

custom = pars['custom']
meta = pars['metadata']
file_data = meta['file_data']
tag = meta['tag']
Ilim = custom['Ilim']

au_to_m = u.au.to('m')
dpc = meta['dpc']*u.pc

datacube = Data(file_data, dpc)
datacube.convert_to_tb(writefits=False)

vchannels = datacube.vchannels

with open('grid_extent.json') as json_file:
    grid = json.load(json_file)

xmax = grid['xsky'] 
extent= np.array([-xmax, xmax, -xmax, xmax])

#**********************************
#READ PARCUBE AND RECONSTRUCT CUBE
#**********************************
if args.kernel in ['gaussian', 'bell']:
    parcube_up = fits.getdata('parcube_%s_data.fits'%args.kernel)
    parcube_low = None
else:
    parcube_up, parcube_low = None, None
    if args.surface in ['upper', 'both']:
        parcube_up = fits.getdata('parcube_up_%s_%s_data.fits'%(args.kernel, args.kind))
    if args.surface in ['lower', 'both']:        
        parcube_low = fits.getdata('parcube_low_%s_%s_data.fits'%(args.kernel, args.kind))

fitdata = fit_kernel.get_channels_from_parcube(parcube_up, parcube_low, vchannels, method=args.kernel, kind=args.kind, n_fit=None)
fitcube = Cube(fitdata, datacube.header, vchannels, dpc, beam=datacube.beam)

#SHOW 
fitcube.show(compare_cubes=[datacube], extent=extent, int_unit='Intensity [K]', vmin=0.0, vmax=3*Ilim, show_beam=True)
fitcube.show_side_by_side(datacube, extent=extent, int_unit='Intensity [K]', vmin=0.0, vmax=3*Ilim, show_beam=True)



from discminer.core import Data
from discminer.disc2d import Model

import numpy as np
import matplotlib.pyplot as plt

from astropy import units as u
import emcee

from argparse import ArgumentParser

parser = ArgumentParser(prog='Handle emcee backend', description='Handle emcee backend')
parser.add_argument('-b', '--backend', default=1, type=int, choices=[0, 1], help="If 0, create new backend. If 1, reuse existing backend")
args = parser.parse_args()

#*********************
#REQUIREd DEFINITIONS
#*********************
file_data = 'MWC_480_CO_220GHz.robust_0.5.JvMcorr.image.pbcor_clipped_downsamp_10pix.fits'
tag_out = 'mwc480_12co_0p2_maps' #PREFERRED FORMAT: disc_mol_chan_program_extratags
tag_in = tag_out

nwalkers = 230
nsteps = 5000

dpc = 162.0*u.pc
vel_sign = -1 #Rotation direction: -1 or 1

Rmax = 1000*u.au #Model maximum radius

#*********
#READ DATA
#*********
datacube = Data(file_data, dpc) #Read data and convert to Cube object
vchannels = datacube.vchannels

au_to_m = u.au.to('m')

#****************************
#INIT MODEL AND PRESCRIPTIONS
#****************************
model = Model(datacube, Rmax, Rmin=0, prototype = False)                  

def intensity_powerlaw_rout(coord, I0=30.0, R0=100*au_to_m, p=-0.4, z0=100*au_to_m, q=0.3, Rout=200):
    if 'R' not in coord.keys(): R = hypot_func(coord['x'], coord['y'])
    else: R = coord['R']
    z = coord['z']
    A = I0*R0**-p*z0**-q
    Ieff = np.where(R<=Rout*au_to_m, A*R**p*np.abs(z)**q, 0.0)
    return Ieff

def z_upper(coord, z0, p, Rb, q, R0=100):
    R = coord['R']/au_to_m
    return au_to_m*(z0*(R/R0)**p*np.exp(-(R/Rb)**q))

def z_lower(coord, z0, p, Rb, q, R0=100):
    R = coord['R']/au_to_m
    return -au_to_m*(z0*(R/R0)**p*np.exp(-(R/Rb)**q))

model.z_upper_func = z_upper
model.z_lower_func = z_lower
model.velocity_func = model.keplerian_vertical # i.e. vrot = sqrt(GM/r**3)*R
model.line_profile = model.line_profile_bell
model.intensity_func = intensity_powerlaw_rout

#If not redefined, intensity and linewidth are powerlaws 
 #of R and z by default, whereas lineslope is constant.
  #See Table 1 of discminer paper 1 (izquierdo+2021).

#****************************
#SET FREE PARS AND BOUNDARIES
#****************************
# If True, parameter is allowed to vary freely.
#  If float, parameter is fixed to the value provided.

model.mc_params['velocity']['vel_sign'] = vel_sign 
model.mc_params['velocity']['Mstar'] = True
model.mc_params['velocity']['vsys'] = True 
model.mc_params['orientation']['incl'] = True
model.mc_params['orientation']['PA'] = True
model.mc_params['orientation']['xc'] = True
model.mc_params['orientation']['yc'] = True
model.mc_params['intensity'] = {'I0': True, 'p': True, 'q': True, 'Rout': True}
model.mc_params['linewidth']['L0'] = True
model.mc_params['linewidth']['p'] = True
model.mc_params['linewidth']['q'] = True
model.mc_params['lineslope']['Ls'] = True
model.mc_params['lineslope']['p'] = True
model.mc_params['height_upper'] = {'z0': True, 'p': True, 'Rb': True, 'q': True}
model.mc_params['height_lower'] = {'z0': True, 'p': True, 'Rb': True, 'q': True}
                                   
# Boundaries of user-defined attributes must be defined here.
# Boundaries of attributes existing in discminer can be modified here, otherwise default values are taken.

model.mc_boundaries['velocity']['vsys'] = (-10, 10)
model.mc_boundaries['orientation']['incl'] = (-1.2, 1.2)
model.mc_boundaries['intensity']['I0'] = (0, 5)
model.mc_boundaries['intensity']['Rout'] = (100, 1000)
model.mc_boundaries['height_upper']['z0'] = (0, 200)
model.mc_boundaries['height_upper']['p'] = (0, 5)
model.mc_boundaries['height_upper']['Rb'] = (0, 1000)
model.mc_boundaries['height_upper']['q'] = (0, 10)
model.mc_boundaries['height_lower']['z0'] = (0, 200)
model.mc_boundaries['height_lower']['p'] = (0, 5)
model.mc_boundaries['height_lower']['Rb'] = (0, 1000)
model.mc_boundaries['height_lower']['q'] = (0, 10)

#***************************
#INITIAL GUESS OF PARAMATERS
#***************************
Mstar = 2.0
vsys = 6.0

incl = -0.8
PA = 0.7
xc = 0
yc = 0

I0 = 1.0
p = -2.5
q = 2.0
Rout = 800

L0 = 0.3
pL = -0.3
qL = -0.4

Ls = 2.0
pLs = 0.3

z0_upper, p_upper, Rb_upper, q_upper = 30, 1.2, 600, 3.0
z0_lower, p_lower, Rb_lower, q_lower = 20, 1.2, 600, 3.0

p0 = [Mstar, vsys,                              #Velocity
      incl, PA, xc, yc,                         #Orientation
      I0, p, q, Rout,                           #Intensity
      L0, pL, qL,                               #Line width
      Ls, pLs,                                  #Line slope
      z0_upper, p_upper, Rb_upper, q_upper,     #Upper surface height
      z0_lower, p_lower, Rb_lower, q_lower      #Lower surface height
]

#********
#RUN MCMC
#********
# Set up the emcee backend
filename = "backend_%s.h5"%tag_out
backend = None

#try and except statement failing with FileNotFoundError/OSError
if args.backend:
    #Succesive runs
    backend = emcee.backends.HDFBackend(filename)
else:
    #First run: Initialise backend
    backend = emcee.backends.HDFBackend(filename)
    backend.reset(nwalkers, len(p0))

print("Backend Initial size: {0} steps".format(backend.iteration))

# Noise in each pixel is stddev of intensity from first and last 5 channels 
noise = np.std( np.append(datacube.data[:5,:,:], datacube.data[-5:,:,:], axis=0), axis=0) 

#Run Emcee
model.run_mcmc(datacube.data, vchannels,
               p0_mean=p0, nwalkers=nwalkers, nsteps=nsteps,
               backend=backend,
               tag=tag_out,
               #nthreads=96, # If not specified considers maximum possible number of cores
               frac_stats=0.1,
               frac_stddev=1e-2,
               noise_stddev=noise) 

print("Backend Final size: {0} steps".format(backend.iteration))



from discminer.core import Data
from utils import add_parser_args

import numpy as np
from astropy import units as u

import os
import json
from argparse import ArgumentParser


parser = ArgumentParser(prog='make moments', description='Make gauss or bell moments and save outputs into .fits files')
args = add_parser_args(parser, kernel=True, sigma=4)

with open('parfile.json') as json_file:
    pars = json.load(json_file)

meta = pars['metadata']

#****************
#SOME DEFINITIONS
#****************
dpc = meta['dpc']*u.pc

#*********
#LOAD DATA
#*********
tag = meta['tag']
file_data = 'cube_data_%s_convtb.fits'%tag
file_model = 'cube_model_%s_convtb.fits'%tag

datacube = Data(file_data, dpc) # Read datacube and convert to Cube object
modelcube = Data(file_model, dpc) # Read modelcube and convert to Cube object

#**********************
#MAKE MOMENT MAPS
#**********************

if args.kernel=='quadratic':
    os.system('bettermoments '+file_data)
    os.system('bettermoments '+file_model)
    for old, new in zip(['_v0', '_dv0', '_Fnu', '_dFnu'], ['velocity_', 'delta_velocity_', 'peakintensity_', 'delta_peakintensity_']):
        os.rename(file_data.split('.fits')[0]+old+'.fits', new+'quadratic_data.fits')
        os.rename(file_model.split('.fits')[0]+old+'.fits', new+'quadratic_model.fits')
    
else:
    moments_data = datacube.make_moments(method=args.kernel, tag='data', sigma_thres=args.sigma)
    moments_model = modelcube.make_moments(method=args.kernel, tag='model', sigma_thres=0)




from discminer.mining_control import _mining_moment_residuals
from discminer.mining_utils import (get_2d_plot_decorators,
                                    init_data_and_model,                                    
                                    get_noise_mask,
                                    load_moments,
                                    load_disc_grid,
                                    mark_planet_location,
                                    show_output)

from discminer.core import Data
from discminer.rail import Contours
from discminer.plottools import (make_up_ax,
                                 mod_major_ticks,
                                 mod_nticks_cbars,
                                 use_discminer_style)

import numpy as np
import matplotlib.pyplot as plt
from astropy import units as u

import copy
import json

use_discminer_style()

if __name__ == '__main__':
    parser = _mining_moment_residuals(None)
    args = parser.parse_args()

#**************************
#JSON AND SOME DEFINITIONS
#**************************    
with open('parfile.json') as json_file:
    pars = json.load(json_file)

meta = pars['metadata']
best = pars['best_fit']
custom = pars['custom']
rings = custom['rings']
vsys = best['velocity']['vsys']
Rout = best['intensity']['Rout']

ctitle, clabel, clim, cfmt, cmap_mom, cmap_res, levels_im, levels_cc, unit = get_2d_plot_decorators(args.moment)

if args.moment=='velocity':
    cmap_mom = copy.copy(cmap_mom)
    cmap_mom.set_under('1')
    cmap_mom.set_over('1')
    
    cmap_res = copy.copy(cmap_res)
    cmap_res.set_under('1')
    cmap_res.set_over('1')
    
#****************
#SOME DEFINITIONS
#****************
file_data = meta['file_data']
tag = meta['tag']

dpc = meta['dpc']*u.pc
Rmax = 1.1*args.Router*Rout*u.au #Max model radius, 10% larger than disc Rout

au_to_m = u.au.to('m')

#********************
#LOAD DATA AND GRID
#********************
datacube, model = init_data_and_model(Rmin=0, Rmax=Rmax)

#Useful definitions for plots
with open('grid_extent.json') as json_file:
    grid = json.load(json_file)

xmax = grid['xsky'] 
xlim = 1.1*np.min([xmax, Rmax.value])
extent= np.array([-xmax, xmax, -xmax, xmax])

#*************************
#LOAD DISC GEOMETRY
R, phi, z = load_disc_grid()
noise_mean, mask = get_noise_mask(datacube, thres=args.sigma,
                                  mask_phi={'map2d': np.degrees(phi['upper']),
                                            'lims': args.mask_phi},
                                  mask_R={'map2d': R['upper']/au_to_m,
                                          'lims': args.mask_R}
)

#*************************
#LOAD MOMENT MAPS    
moment_data, moment_model, residuals, mtags = load_moments(args, mask=mask)

#**************************
#MAKE PLOT
fig, ax = plt.subplots(ncols=3, nrows=1, figsize=(15,6))
ax_cbar0 = fig.add_axes([0.15, 0.14, 0.450, 0.04])
ax_cbar2 = fig.add_axes([0.68, 0.14, 0.212, 0.04])
    
kwargs_im = dict(cmap=cmap_mom, extent=extent, levels=levels_im)
kwargs_cc = dict(colors='k', linestyles='-', extent=extent, levels=levels_cc, linewidths=0.4)
kwargs_cbar = dict(orientation='horizontal', pad=0.03, shrink=0.95, aspect=15)

im0 = ax[0].contourf(moment_data, extend='both', **kwargs_im)
im1 = ax[1].contourf(moment_model, extend='both', **kwargs_im)
im2 = ax[2].contourf(residuals, cmap=cmap_res, origin='lower', extend='both', extent=extent, levels=np.linspace(-1.01*clim, 1.01*clim, 32))

cc0 = ax[0].contour(moment_data, **kwargs_cc)
cc1 = ax[1].contour(moment_model, **kwargs_cc)

cbar0 = plt.colorbar(im0, cax=ax_cbar0, format='%.1f', **kwargs_cbar)
cbar0.ax.tick_params(labelsize=12) 
cbar2 = plt.colorbar(im2, cax=ax_cbar2, format=cfmt, **kwargs_cbar)
cbar2.ax.tick_params(labelsize=12) 

mod_nticks_cbars([cbar0], nbins=10)
mod_nticks_cbars([cbar2], nbins=5)

ax[0].set_ylabel('Offset [au]', fontsize=15)
ax[0].set_title(ctitle, pad=40, fontsize=17)
ax[1].set_title('Discminer Model', pad=40, fontsize=17)
ax[2].set_title('Residuals', pad=40, fontsize=17)

cbar0.set_label(clabel, fontsize=14)
cbar2.set_label(r'Residuals %s'%unit, fontsize=14)

for i,axi in enumerate(ax):
    make_up_ax(axi, xlims=(-xlim, xlim), ylims=(-xlim, xlim), labelsize=11)
    mod_major_ticks(axi, axis='both', nbins=8)
    axi.set_aspect(1)

    if i>=1:
        axi.tick_params(labelleft=False)
        
    Contours.emission_surface(axi, R, phi, extent=extent, R_lev=np.arange(25, 0.98*Rout, 50)*u.au.to('m'), which='both')
    r"""
    #Overlay dust rings?
    Contours.emission_surface(
        axi, R, phi, extent=extent,
        R_lev=np.array(rings)*u.au.to('m'), which='upper',
        kwargs_R={'linestyles': '-', 'linewidths': 1.2, 'colors': 'k'},
        kwargs_phi={'colors': 'none'}
    )
    #"""
    
datacube.plot_beam(ax[0], fc='0.8')
mark_planet_location(ax[2], args, edgecolor='k', lw=2.0, s=200, coords='sky', model=model, midplane=False)

plt.savefig('moment+residuals_%s.png'%mtags['base'], bbox_inches='tight', dpi=200)
show_output(args)




from discminer.mining_control import _mining_radial_profiles_wedge
from discminer.mining_utils import (init_data_and_model,
                                    get_noise_mask,
                                    get_1d_plot_decorators,
                                    load_moments,
                                    load_disc_grid,
                                    show_output,
                                    MEDIUM_SIZE)

from discminer.rail import Rail
from discminer.plottools import (get_discminer_cmap,
                                 make_substructures,
                                 make_up_ax,
                                 mod_minor_ticks,
                                 mod_major_ticks,
                                 mod_nticks_cbars,
                                 make_1d_legend,
                                 truncate_colormap,
                                 use_discminer_style)
                                 
import json
import numpy as np
from astropy import units as u
from matplotlib import patches
import matplotlib.pyplot as plt
from scipy.signal import savgol_filter

use_discminer_style()

if __name__ == '__main__':
    parser = _mining_radial_profiles_wedge(None)
    args = parser.parse_args()

#**************************
#JSON AND SOME DEFINITIONS
#**************************
with open('parfile.json') as json_file:
    pars = json.load(json_file)

meta = pars['metadata']
best = pars['best_fit']
custom = pars['custom']

Mstar = best['velocity']['Mstar']
vsys = best['velocity']['vsys']
vel_sign = best['velocity']['vel_sign']
Rout = best['intensity']['Rout']
incl = best['orientation']['incl']

gaps = custom['gaps']
rings = custom['rings']
Rmax = 1.1*Rout*u.au #Max model radius, 10% larger than disc Rout

clabel, clabel_res, clim0, clim0_res, clim1, clim1_res, unit = get_1d_plot_decorators(args.moment)

#*******************
#LOAD DATA AND MODEL
#*******************
datacube, model = init_data_and_model()

noise_mean, mask = get_noise_mask(datacube, thres=2)
vchannels = datacube.vchannels
model.make_model()

#*************************
#LOAD DISC GEOMETRY
R, phi, z = load_disc_grid()

R_s = R[args.surface]*u.m.to('au')
phi_s = phi[args.surface]

#*************************
#LOAD MOMENT MAPS
moment_data, moment_model, residuals, mtags = load_moments(args, mask=mask)
ref_surf = mtags['ref_surf']
tag_base = mtags['base']

#*************************
#ABSOLUTE RESIDUALS    
if args.moment=='velocity':
    residuals_abs = np.abs(moment_data-vsys) - np.abs(moment_model-vsys)

#*************************
#RADIAL BINS AND UTILS
beam_au = datacube.beam_size.to('au').value
if args.absolute_Rinner>=0:
    Rprof_in = args.absolute_Rinner
else:
    Rprof_in = args.Rinner*beam_au

if args.absolute_Router>=0:
    Rprof_out = args.absolute_Router
else:
    Rprof_out = args.Router*Rout
    
R_prof = np.arange(Rprof_in, Rprof_out, beam_au/4.0) #changed to beam/4 from beam/5 before
xlim0, xlim1 = 0.5*R_prof[0], 1.05*R_prof[-1]

def make_zones_code(fig, ax, az_zones, color_zones):
    figx, figy = fig.get_size_inches()
    figr = figy/figx
    axp = ax.get_position()    
    dy = 0.3
    ax1 = fig.add_axes([axp.x1 - dy*figr/2., axp.y1 - dy/2, dy*figr, dy])    
    ax1.set_xlim(-1.1, 1.1)
    ax1.set_ylim(-1.1, 1.1)

    wedges = []
    for i,zone in enumerate(az_zones):
        wedges.append(patches.Wedge((0,0), 1, zone[0], zone[1], hatch=None, fc=color_zones[str(i)], fill=True, alpha=0.8))

    circled = patches.Circle((0,0), 1, lw=3, ec='k', fill=False)
    circlef = patches.Circle((0,0), 1, fc='w', ec='none', fill=True)    

    ax1.add_artist(circlef)

    for wedge in wedges:
        ax1.add_artist(wedge)

    ax1.add_artist(circled)
    ax1.axis('off')

    return ax1

def make_savgol(prof):
    if not args.savgol_filter:
        return prof, None
    try:
        ysav = savgol_filter(prof, 5, 3)
        ysav_deriv = savgol_filter(prof, 5, 3, deriv=1)
    except np.linalg.LinAlgError:
        ysav = prof
        ysav_deriv = None
    return ysav, ysav_deriv 

#**********************************
#AZIM. AVERAGED RESIDUALS PER ZONE
#**********************************
rail = Rail(model, residuals, R_prof)
az_zones = [[a, b] for a,b in zip(args.wedges[::2], args.wedges[1::2])]
cmap = truncate_colormap(plt.get_cmap('nipy_spectral'), minval=0.1, maxval=0.9)
color_list = [cmap(ind) for ind in np.linspace(0, 1, len(az_zones))]
color_zones = {str(i): color_list[i] for i in range(len(az_zones))} #{'0': '#d90429', '1': '#f77f00', '2': '#43ABC9', '3': '#0d3d56'}
zones, zones_error  = rail.get_average_zones(az_zones=az_zones, sigma_thres=args.sigma)

fig, ax = plt.subplots(ncols=1, nrows=1, figsize=(11, 3))
ax1 = make_zones_code(fig, ax, az_zones, color_zones)

for i in range(len(az_zones)):
    try:
        ysav, ysav_deriv = make_savgol(zones[i])
        ax.plot(R_prof, ysav, c=color_zones[str(i)], lw=2, zorder=11)
        ax.fill_between(R_prof, zones[i]+zones_error[i], zones[i]-zones_error[i], color=color_zones[str(i)], alpha=0.15, zorder=9) #, label='zone '+str(i))
    except np.linalg.LinAlgError:
        ax.scatter(R_prof, zones[i], c=color_zones[str(i)], edgecolor='0.1', zorder=11)
            
for i,lev in enumerate(R_prof):
    ax.scatter([lev]*len(rail._resid_list[i]), rail._resid_list[i], color='0.9', alpha=0.9, s=5, zorder=-1)
ax.scatter(None, None, color='0.8', alpha=0.9, s=10, zorder=-1, label='Points from all azimuths')    

make_up_ax(ax, labelbottom=True, labeltop=False, xlims=(R_prof[0], R_prof[-1]), ylims=np.array([clim0_res, clim1_res]))
mod_major_ticks(ax, axis='x', nbins=8)
mod_major_ticks(ax, axis='y', nbins=5)

make_substructures(ax, gaps=gaps, rings=rings, label_gaps=True, label_rings=True)      

ax.set_xlabel('Radius [au]')
if args.moment in ['velocity', 'v0phi', 'v0r', 'v0z', 'vr_leftover']:
    ax.set_ylabel(r'$\delta\upsilon$ [km/s]', fontsize=MEDIUM_SIZE, labelpad=10)    

make_1d_legend(ax, scatterpoints=3, fontsize=13, loc='lower right', bbox_to_anchor=(0.9,1.0))

plt.savefig('wedge_radprof_residuals_%s.png'%tag_base, bbox_inches='tight', dpi=200)
show_output(args)



"""
Pick module
===========
Classes: Pick
"""

import numpy as np

from sklearn.cluster import KMeans
from sklearn.metrics import silhouette_score
import scipy.stats as st
from .rail import Rail
from .tools.utils import weighted_std

import copy
import warnings

def get_neighbour_peaks(var_x, pos_x, var_y, n_clusters=8, detect_thres=3):
    #Everything is referred to the variance sorted wrt x
    ind_x_sort = np.argsort(pos_x)
    var_x = var_x[ind_x_sort]
    pos_x = pos_x[ind_x_sort]
    var_y = var_y[ind_x_sort]
    ind_main = np.argmax(var_y)
    var_nomain = np.delete(var_y, ind_main)
    acc_neigh = []
    delete_list = [ind_main]
    ind_centre = ind_main
    first_neighs = np.sort(var_y[np.abs(np.arange(0, n_clusters) - ind_main)==1])[::-1]

    while(True):
        ind_neigh, = np.where(var_y==first_neighs[0])[0]
        delete_list.append(ind_neigh)
        var_noneigh = np.delete(var_y, delete_list)
        var_std_noneigh = np.std(var_noneigh)
        var_mean_noneigh = np.mean(var_noneigh)
        if first_neighs[0] >= var_mean_noneigh+detect_thres*var_std_noneigh:
            acc_neigh.append(ind_neigh)
            sign = int(np.sign(ind_neigh-ind_main)) #left or right neighbour?
            if len(first_neighs)<2: first_neighs = np.array([var_y[ind_neigh+sign]]) #If already at a border
            elif ind_neigh+sign>=n_clusters or ind_neigh+sign<0: first_neighs = np.array([var_y[ind_centre-sign]]) #If neighbour out of the border
            else:
                first_neighs = np.sort([var_y[ind_neigh+sign], var_y[ind_centre-sign]])[::-1] #else if exactly two neighbours
                ind_centre, = np.where(var_y==first_neighs[1])[0]
                ind_centre = ind_centre - int(np.sign(ind_centre-ind_main)) #If the smallest neigh is on the right of ind_main, then ind_centre=ind_smallest-1
        else: break #if first_neighs[i] < 3sigma then the other peak is also smaller

    if len(acc_neigh)==0:
        var_std = np.std(var_nomain)
        var_mean = np.mean(var_nomain)
    else:
        var_std = var_std_noneigh
        var_mean = var_mean_noneigh

    acc_neigh = np.asarray(acc_neigh).astype(int)
    pos_left = pos_x[ind_main]
    pos_right = pos_x[ind_main]
    var_left, var_right = var_x[ind_main], var_x[ind_main]
    if var_y[ind_main] > var_mean+detect_thres*var_std:
        acc_neigh = np.append(acc_neigh, ind_main)
    if len(acc_neigh)>1:
        ind_left = np.argmin(pos_x[acc_neigh])
        ind_right = np.argmax(pos_x[acc_neigh])
        pos_left = pos_x[acc_neigh[ind_left]]
        var_left = var_x[acc_neigh[ind_left]]
        pos_right = pos_x[acc_neigh[ind_right]]
        var_right = var_x[acc_neigh[ind_right]]

    #accepted peaks are sorted from lowest to highest peak. The returned indices refer to the list of peaks sorted wrt the input x coord
    return acc_neigh, var_std, [pos_left, var_left, pos_right, var_right]


class Pick(Rail):
    def __init__(self, model, prop, coord_levels, **kw_prop_along_coords):
        super().__init__(model, prop, coord_levels)
        kw_pac = dict(fold=True)
        kw_pac.update(kw_prop_along_coords)
        lev, coord, resid, color = self.prop_along_coords(**kw_pac)

        self.lev_list = copy.copy(lev)        
        self.coord_list = copy.copy(coord)
        self.resid_list = copy.copy(resid)
        self.color_list = copy.copy(color)

    def find_peaks(self, phi_min=-85, phi_max=85, detect_thres=3, clean_thres=np.inf, clean_histogram=True, fig_ax_histogram=None, av_global=np.median, init_thres=5.0, init_std=0.2):
        len_res = len(self.resid_list)
        peak_angle = np.zeros(len_res)
        peak_resid = np.zeros(len_res)
        peak_sign = np.zeros(len_res)
        peak_error = np.zeros(len_res)
        
        for i in np.arange(0, len_res): #1 radius per quadrant (0,90), (-90,0)
            arg90 = (self.coord_list[i] >= phi_min) & (self.coord_list[i] <= phi_max)  #-90, 90
            if np.sum(arg90) == 0: abs_resid = [np.nan] #if arg90 is empty
            else: abs_resid = np.abs(self.resid_list[i][arg90])

            argpeak = np.argmax(abs_resid)
            peak_resid[i] = abs_resid[argpeak]

            if np.sum(arg90) == 0: #if arg90 is empty
                peak_angle[i] = np.nan
                peak_sign[i] = np.nan
            else:
                peak_angle[i] = self.coord_list[i][arg90][argpeak]
                peak_sign[i] = np.sign(self.resid_list[i][arg90][argpeak])

        if clean_thres is not None and np.isfinite(clean_thres):

            got_runtimeerror = False
            
            if clean_histogram:

                from discminer.tools.fit_kernel import _gauss
                from scipy.optimize import curve_fit

                yfunc = lambda x, A, sigma: _gauss(x, A, 0, sigma)
                
                peak_hist = peak_resid[peak_resid < init_thres] #Initial threshold of 5 km/s, not suited for folded intensities
                counts, bins = np.histogram(peak_hist, bins=4*int(round(len(peak_resid)**(1/3.)))-1 )
                mbins = 0.5*(bins[1:] + bins[:-1])
                
                try:
                    popt, pcov = curve_fit(yfunc, mbins, counts, p0=[np.max(counts), init_std])
                    popt = np.abs(popt) #Make sure A and sigma are positive, as our residuals are

                    if fig_ax_histogram is not None:
                        
                        import matplotlib.pyplot as plt

                        fig, ax = fig_ax_histogram
                        ax.stairs(counts, bins, color='dodgerblue', lw=3)
                    
                        xgauss = np.linspace(0, round(np.nanmax(bins)), 100)
                        ygauss = yfunc(xgauss, *popt)
                    
                        ax.plot(xgauss, ygauss, lw=3, c='k')
                        ax.axvline(clean_thres*np.abs(popt[1]), lw=4, c='tomato')
                    
                        ax.set_xlabel('Peak residual [km/s]')
                        ax.set_ylabel('Counts')

                    ii = peak_resid < clean_thres*popt[1]
                    print ('Rejecting %d peak residuals above %.3f (%dsigma)'%(np.sum(~ii), clean_thres*popt[1], clean_thres))
                    self.cutout_thres = clean_thres*popt[1]
                    
                except RuntimeError as e:
                    print('RuntimeError:', e, '*******Rejection of points through histogram method failed. Trying median(peaks)+thres*sigma(peaks) rejection...')
                    got_runtimeerror = True
                    
            if not clean_histogram or got_runtimeerror:
                rej_thresh = np.nanmedian(peak_resid) + clean_thres*np.nanstd(peak_resid)
                ii = peak_resid < rej_thresh
                print ('Rejecting %d peak velocity residuals above %.3f km/s (median+%dsigma)'%(np.sum(~ii), rej_thresh, clean_thres))
                self.cutout_thres = rej_thresh
                
            self.lev_list = self.lev_list[ii]
            self.color_list = self.color_list[ii]
            self.coord_list = self.coord_list[ii]
            self.resid_list = self.resid_list[ii]
            peak_resid = peak_resid[ii]
            peak_angle = peak_angle[ii]
            peak_sign = peak_sign[ii]
            
        #peak_sky_coords = get_sky_from_disc_coords(self.lev_list, np.radians(peak_angle))
        peak_error = np.ones(len(self.lev_list)) #np.array([centroid_errors2d(peak_sky_coords[0][i], peak_sky_coords[1][i])[0] for i in range(len_res)])
        #peak_error[peak_error/peak_resid>2] = 0
        peak_weight = np.where(peak_error==0, 0, 1/peak_error)

        #********************
        #FIND GLOBAL PEAK
        #********************
        peak_mean = np.nansum(peak_weight*peak_resid)/np.nansum(peak_weight)
        peak_std = weighted_std(peak_resid, peak_weight, weighted_mean=peak_mean)
        for stdi in np.arange(detect_thres+1)[::-1]: #If it fails to find peaks above the threshold continue and try the next, lower, sigma threshold
            ind_global_peak = peak_resid > peak_mean+stdi*peak_std
            if np.nansum(ind_global_peak) > 0:
                print ('Global peak residual, dv=%.2f, found above %d sigma from mean value...'%(np.max(peak_resid), stdi))                
                break

        """
        self.peak_global_val = np.max(peak_resid[ind_global_peak])
        self.peak_global_angle = np.median(peak_angle[ind_global_peak])
        self.peak_global_radius = np.median(self.lev_list[ind_global_peak])
        """

        #Mean and std from all points
        self.peak_mean = peak_mean
        self.peak_std = peak_std
        
        #Global peak into
        ind_max = np.argmax(peak_resid[ind_global_peak])
        self.peak_global_val = peak_resid[ind_global_peak][ind_max]
        self.peak_global_sign = peak_sign[ind_global_peak][ind_max]
        self.peak_global_angle = peak_angle[ind_global_peak][ind_max]
        self.peak_global_radius = self.lev_list[ind_global_peak][ind_max]
        self.peak_global_sigma = (self.peak_global_val-peak_mean)/peak_std
        
        #Location, amplitude and other properties, for all points
        nonan = ~np.isnan(peak_resid)
        
        self.peak_resid = peak_resid[nonan]
        self.peak_angle = peak_angle[nonan]
        self.peak_sign = peak_sign[nonan]
        self.peak_error = peak_error[nonan]
        self.peak_weight = peak_weight[nonan]

        self.lev_list = self.lev_list[nonan]        
        self.color_list = self.color_list[nonan]
        self.coord_list = self.coord_list[nonan]
        self.resid_list = self.resid_list[nonan]
        

    def _make_clusters(self, n_clusters, axis='phi'): 

        #************************
        #APPLY K-MEANS CLUSTERING
        #************************
        n_peaks = len(self.peak_resid)
        if axis=='phi': peak_both = np.array([self.peak_angle, self.peak_resid]).T
        elif axis=='R': peak_both = np.array([self.lev_list, self.peak_resid]).T
        kmeans = KMeans(n_clusters=n_clusters, random_state=0).fit(peak_both)
        kcenters = kmeans.cluster_centers_
        klabels = kmeans.labels_

        npoints_clust = np.zeros(n_clusters)
        for i in range(n_clusters): npoints_clust[i]=np.sum(klabels==i)

        ind_cluster_peak = np.argmax(kcenters[:,1])
        peak_cluster_nomax = np.sort(kcenters[:,1])[:-1]
        peak_cluster_mean = np.mean(peak_cluster_nomax)
        peak_cluster_std = np.std(peak_cluster_nomax)
        peak_cluster_angle = kcenters[:,0][ind_cluster_peak]

        #**********************
        #FIND CLUSTER VARIANCES
        #**********************
        variance_x, variance_y = [], []
        percentiles = [5, 67] #95%, 68% and 33% contours
        #kde_color = '#FFB000'
        #kde_cmap = get_cmap_from_color(kde_color, lev=len(percentiles)) #'#DC267F' #'#FFB000' #'#648FFF'

        if axis=='phi': xmin_c, xmax_c = -90*1.05, 90*1.05
        elif axis=='R': xmin_c, xmax_c = np.nanmin(self.lev_list), np.nanmax(self.lev_list)
        ymin_c, ymax_c = -0.3*self.peak_global_val, 1.5*self.peak_global_val
        xx, yy = np.mgrid[xmin_c:xmax_c:200j, ymin_c:ymax_c:200j]

        for i in range(n_clusters):
            x = peak_both[:,0][klabels == i]
            y = peak_both[:,1][klabels == i]
            positions = np.vstack([xx.ravel(), yy.ravel()])
            values = np.vstack([x, y]) #shape (2, ndata)
            try:
                kernel = st.gaussian_kde(values)#, weights=peak_weight[klabels==i])
                perr = np.sqrt(np.diag(kernel.covariance))
                variance_x.append((kernel.covariance[0,0]**0.5/kernel.factor)**2) #True variance: std**2
                variance_y.append(kernel.covariance[1,1])
            except: #when too few points in cluster
                variance_x.append(0)
                variance_y.append(0)
                continue

            f = np.reshape(kernel(positions).T, xx.shape)
            levels = [st.scoreatpercentile(kernel(kernel.resample(1000)), per) for per in percentiles]
            levels.append(np.amax(f))

        #********************
        variance_x = np.asarray(variance_x)
        variance_y = np.asarray(variance_y) * npoints_clust*(n_clusters/n_peaks) #weighting by number of points in clusters
        if axis=='phi': self.klabels_phi = klabels
        if axis=='R': self.klabels_R = klabels
        return kcenters, variance_x, variance_y

    def find_clusters(self, n_phi=8, n_R=8, detect_thres=3):

        kcenters_phi, variance_phi_x, variance_phi_y = self._make_clusters(n_phi, axis='phi')
        kcenters_R, variance_R_x, variance_R_y = self._make_clusters(n_R, axis='R')

        #***************************
        #FIND PEAK AZIMUTHAL CLUSTER VARIANCES
        #***************************
        acc_peaks_phi, var_std_phi, var_width_phi = get_neighbour_peaks(variance_phi_x, kcenters_phi[:,0], variance_phi_y, n_clusters=n_phi, detect_thres=detect_thres)
        print ("accepted variance peaks on PHI:", acc_peaks_phi)

        var_colors_phi = np.array(['1.0']*n_phi).astype('<U9')
        var_colors_phi[acc_peaks_phi] = 'palegreen' #kde_color
        ind_variance_peak_phi = np.argmax(variance_phi_y)
        variance_nomax = np.sort(variance_phi_y)[:-len(acc_peaks_phi)]
        peak_variance_std_phi = var_std_phi #std of variances except those from accepted peaks np.std(variance_nomax)
        self.peak_variance_phi = kcenters_phi[:,0][ind_variance_peak_phi]

        kc_indsort_phi = np.argsort(kcenters_phi[:,0])
        kcent_sort_phi = kcenters_phi[:,0][kc_indsort_phi]
        kcent_sort_vel_phi = kcenters_phi[:,1][kc_indsort_phi]
        var_y_sort_phi = variance_phi_y[kc_indsort_phi]
        self.var_nopeaks_phi = np.delete(var_y_sort_phi, acc_peaks_phi)
        peak_variance_mean_phi = np.mean(self.var_nopeaks_phi) #mean of variances excluding those from accepted peaks

        #peak_variance_sigmas = (self.variance_y[ind_variance_peak]-peak_variance_mean)/peak_variance_std #Considering only peak accepted variance
        self.peak_variance_sigmas_phi_mean = (np.mean(var_y_sort_phi[acc_peaks_phi])-peak_variance_mean_phi)/peak_variance_std_phi #Considering mean std of all accepted peaks
        self.peak_variance_sigmas_phi = (var_y_sort_phi[acc_peaks_phi]-peak_variance_mean_phi)/peak_variance_std_phi #std of each accepted peak        
        #***************************
        #FIND PEAK RADIAL CLUSTER VARIANCES
        #***************************
        acc_peaks_R, var_std_R, var_width_R = get_neighbour_peaks(variance_R_x, kcenters_R[:,0], variance_R_y, n_clusters=n_R, detect_thres=detect_thres)
        print ("accepted variance peaks on R:", acc_peaks_R)

        var_colors_R = np.array(['1.0']*n_R).astype('<U9')
        var_colors_R[acc_peaks_R] = 'palegreen' #kde_color
        ind_variance_peak_R = np.argmax(variance_R_y)
        variance_nomax = np.sort(variance_R_y)[:-len(acc_peaks_R)]
        peak_variance_std_R = var_std_R #std of variances except those from accepted peaks np.std(variance_nomax)
        self.peak_variance_R = kcenters_R[:,0][ind_variance_peak_R]

        kc_indsort_R = np.argsort(kcenters_R[:,0])
        kcent_sort_R = kcenters_R[:,0][kc_indsort_R]
        kcent_sort_vel_R = kcenters_R[:,1][kc_indsort_R]
        var_y_sort_R = variance_R_y[kc_indsort_R]
        self.var_nopeaks_R = np.delete(var_y_sort_R, acc_peaks_R)
        peak_variance_mean_R = np.mean(self.var_nopeaks_R) #mean of variances excluding those from accepted peaks

        #peak_variance_sigmas = (self.variance_y[ind_variance_peak]-peak_variance_mean)/peak_variance_std #Considering only peak accepted variance
        self.peak_variance_sigmas_R_mean = (np.mean(var_y_sort_R[acc_peaks_R])-peak_variance_mean_R)/peak_variance_std_R #Considering mean std of all accepted peaks
        self.peak_variance_sigmas_R = (var_y_sort_R[acc_peaks_R]-peak_variance_mean_R)/peak_variance_std_R #std of each accepted peak        

        #**************************
        self.acc_peaks_phi = acc_peaks_phi
        self.acc_peaks_R = acc_peaks_R
        
        self.kcent_sort_phi = kcent_sort_phi
        self.var_y_sort_phi = var_y_sort_phi

        self.kcent_sort_R = kcent_sort_R
        self.var_y_sort_R = var_y_sort_R

        self.var_colors_phi = var_colors_phi
        self.var_colors_R = var_colors_R

        self.kc_indsort_phi = kc_indsort_phi
        self.kc_indsort_R = kc_indsort_R

        self.kcent_sort_vel_R = kcent_sort_vel_R
        self.kcent_sort_vel_phi = kcent_sort_vel_phi

        #Background properties
        self.peak_variance_std_R = peak_variance_std_R
        self.peak_variance_mean_R = peak_variance_mean_R
        self.peak_variance_std_phi = peak_variance_std_phi
        self.peak_variance_mean_phi = peak_variance_mean_phi
        
        #Weighted mean location of accepted clusters
        self.acc_phi = np.sum(kcent_sort_phi[acc_peaks_phi] * self.peak_variance_sigmas_phi) / np.sum(self.peak_variance_sigmas_phi)                
        self.acc_R = np.sum(kcent_sort_R[acc_peaks_R] * self.peak_variance_sigmas_R) / np.sum(self.peak_variance_sigmas_R)

    def writetxt(self, filename='pick_summary.txt'):

        ncols = 7

        #Global peak
        try:
            arr_global_peak = [
                [
                    self.peak_global_angle,
                    self.peak_global_radius,
                    self.peak_global_val*self.peak_global_sign,
                    self.peak_global_sigma,
                    self.peak_mean,
                    self.peak_std,
                    'global_peak'
                ]
            ]
        except AttributeError:
            arr_global_peak = [[np.nan]*ncols]
            print ('Writing .txt pick summary file without global peak information...')            

        #Clusters
        arr_clusters_phi = []
        arr_clusters_R = []
        arr_clusters_acc = []        
        try:
            for n, ind in enumerate(self.acc_peaks_phi):
                arr_clusters_phi.append(
                    [
                        self.kcent_sort_phi[ind],
                        np.nan,
                        self.var_y_sort_phi[ind],
                        self.peak_variance_sigmas_phi[n],
                        self.peak_variance_mean_phi,
                        self.peak_variance_std_phi,
                        'cluster_phi_%d'%n
                    ]
                )


            for n, ind in enumerate(self.acc_peaks_R):
                arr_clusters_R.append(
                    [                    
                        np.nan,
                        self.kcent_sort_R[ind],
                        self.var_y_sort_R[ind],
                        self.peak_variance_sigmas_R[n],
                        self.peak_variance_mean_R,
                        self.peak_variance_std_R,
                        'cluster_R_%d'%n
                    ]
                )
            
            arr_clusters_acc = [
                [
                    self.acc_phi,
                    self.acc_R,
                    np.nan,
                    np.nan,
                    np.nan,
                    np.nan,                
                    'weighted_mean_centre_from_accepted_clusters'
                ]
            ]
        except AttributeError:
            arr_clusters_phi.append([np.nan]*ncols)
            arr_clusters_R.append([np.nan]*ncols)
            arr_clusters_acc.append([np.nan]*ncols)
            print ('Writing .txt pick summary file without cluster information...')

        #Gather info
        arr_tot = np.asarray(arr_global_peak + arr_clusters_phi + arr_clusters_R + arr_clusters_acc, dtype=object).squeeze()

        np.savetxt(
            filename,
            arr_tot,
            fmt=('%.2f %.2f %.2e %.2f %.2e %.2e %s').split(),
            header='phi[deg]\tR[au]\tvalue\tsigma\tmean_bckg\tstd_bckg\tcomments',
            delimiter='\t'
        )
        



from .tools.utils import FrontendUtils
from astropy import units as u
from scipy.optimize import root
import numpy as np

_break_line = FrontendUtils._break_line
au_to_m = u.au.to('m')

def grid(xmax, nx, indexing="xy", verbose=True):
    """
    Compute Cartesian (x,y) and polar (R, phi) grid. Assuming square grid, namely ymax=xmax, and ny=nx.
    
    Parameters
    ----------
    xmax : `~astropy.units.Quantity`
        Maximum spatial extent of the grid. 
        The 2D extent of the output grid is [-xmax, xmax] in x and [-xmax, xmax] in y.
    nx : int
        Number of grid cells along each spatial dimension.
    indexing : str, optional
        Cartesian (‘xy’, default) or matrix (‘ij’) indexing of output xy meshgrid. See `~numpy.meshgrid`.
    verbose : bool, optional
        if True, print informative messages.

    Returns
    -------    
    grid : dict
        Dictionary containing grid information. Cartesian and polar grids are returned in units of metres.
    
    Examples
    --------        

    """
    _break_line()
    xmax = xmax.to(u.m).value
    xymax = np.array([xmax, xmax])
    nx = np.int32(nx)
    step = 2 * xmax / (nx - 1)

    if verbose:
        print("Computing grid...")
        print("Grid maximum extent:", xmax)
        print("Grid step (cell size):", step)

    xgrid = np.linspace(-xmax, xmax, nx)
    xygrid = [xgrid, xgrid]
    XY = np.meshgrid(xgrid, xgrid, indexing=indexing)
    xList, yList = [xy.flatten() for xy in XY]
    RList = np.linalg.norm([xList, yList], axis=0)
    phiList = np.arctan2(yList, xList)
    #phiList = np.where(phiList < 0, phiList + 2 * np.pi, phiList)
    _break_line()

    return {
        "x": xList,
        "y": yList,
        "R": RList,
        "phi": phiList,
        "meshgrid": XY,
        "xygrid": xygrid,
        "nx": nx,
        "xmax": xmax,
        "ncells": nx ** 2,
        "step": step,
        "extent": np.array([-xmax, xmax, -xmax, xmax])*u.m.to(u.au)
    }

def grid_weighted(prop, X, Y, Rmin=0*u.au, Rmax=1000*u.au, norm=None, power=3, npoints=5000, fcond=lambda a, b, c: True):
    
    ni, nj = X.shape
    X_au, Y_au = X.to('au').value, Y.to('au').value
    Rmin_au, Rmax_au = Rmin.to('au').value, Rmax.to('au').value
    x, y, values = np.zeros((3, npoints))

    if norm is None:
        norm = np.nanmax(prop)
    
    def accept_point(val):
        flag = np.random.random()
        val = (val / norm)**power
        if val >= flag:
            return True
        else:
            return False

    n = 0        
    while (n < npoints):        
        i = np.random.randint(ni)
        j = np.random.randint(nj)
        val = prop[i,j]
        xi = X_au[i,j]
        yi = Y_au[i,j]

        if accept_point(val) and Rmin_au<=np.hypot(xi,yi)<=Rmax_au and fcond(val, xi, yi):
            x[n] = xi
            y[n] = yi
            values[n] = val
            n+=1
        else:
            continue
    
    return x, y, values


class GridTools:
    @staticmethod
    def _rotate_sky_plane(x, y, ang):
        xy = np.array([x,y])
        cos_ang = np.cos(ang)
        sin_ang = np.sin(ang)
        rot = np.array([[cos_ang, -sin_ang],
                        [sin_ang, cos_ang]])
        return np.dot(rot, xy)

    @staticmethod    
    def _rotate_sky_plane_ewise(x, y, ang):
        cos_ang = np.cos(ang)
        sin_ang = np.sin(ang)
        x_rot = cos_ang * x - sin_ang * y
        y_rot = sin_ang * x + cos_ang * y
        return x_rot, y_rot
    
    @staticmethod
    def _rotate_sky_plane3d(x, y, z, ang, axis='z'):
        xyz = np.array([x,y,z])
        cos_ang = np.cos(ang)
        sin_ang = np.sin(ang)
        if axis == 'x':
            rot = np.array([[1, 0, 0],
                            [0, cos_ang, -sin_ang],
                            [0, sin_ang, cos_ang]])
        if axis == 'y':
            rot = np.array([[cos_ang, 0, -sin_ang],
                            [0, 1, 0],
                            [sin_ang, 0, cos_ang]])
            
        if axis == 'z':
            rot = np.array([[cos_ang, -sin_ang , 0],
                            [sin_ang, cos_ang, 0], 
                            [0, 0, 1]])
        return np.dot(rot, xyz)

    @staticmethod
    def _project_on_skyplane(x, y, z, cos_incl, sin_incl):
        x_pro = x
        y_pro = y * cos_incl - z * sin_incl
        z_pro = y * sin_incl + z * cos_incl
        return x_pro, y_pro, z_pro

    @staticmethod
    def get_sky_from_disc_coords(R, az, z, incl, PA, xc=0, yc=0):
        xp = R*np.cos(az)
        yp = R*np.sin(az)
        zp = z
        xp, yp, zp = GridTools._project_on_skyplane(xp, yp, zp, np.cos(incl), np.sin(incl))
        if len(np.atleast_1d(PA)) > 0:
            xp, yp = GridTools._rotate_sky_plane_ewise(xp, yp, PA)
        else:
            xp, yp = GridTools._rotate_sky_plane(xp, yp, PA)
        return xp+xc, yp+yc, zp

    @staticmethod
    def get_disc_from_sky_coords(xs, ys, z_func, z_pars, incl, PA, xc=0, yc=0, midplane=False):
        #xs, ys: x and y on sky plane
        if len(np.atleast_1d(PA)) > 0:
            xs, ys = GridTools._rotate_sky_plane_ewise(xs-xc, ys-yc, -PA)
        else:
            xs, ys = GridTools._rotate_sky_plane(xs-xc, ys-yc, -PA)
        xd = xs
        cos_incl = np.cos(incl)
        sin_incl = np.sin(incl)
        def find_yd(yd):
            R = np.sqrt(xd**2+yd[0]**2)
            if midplane: zd = 0
            else: zd = z_func({'R': R*au_to_m}, **z_pars)/au_to_m
            return yd[0]*cos_incl - zd*sin_incl - ys #see _project_on_skyplane() 
        yd = root(find_yd, [100], method='hybr')
        return xd, yd.x[0]
    
